/* streamexs.c
 *
 * Based on rstream.c by Sean Hefty.
 *
 * Copyright (c) 2011-2012 Intel Corporation.  All rights reserved.
 * Copyright (c) 2014 University of New Hampshire InterOperability Laboratory
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AWV
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <exs.h>
#include "common.h"
#include "utilities.h"

struct test_size_param {
	int size;
	int option;
};

static struct test_size_param test_size[] = {
	{ 1 <<  6, 0 },
	{ 1 <<  7, 1 }, { (1 <<  7) + (1 <<  6), 1},
	{ 1 <<  8, 1 }, { (1 <<  8) + (1 <<  7), 1},
	{ 1 <<  9, 1 }, { (1 <<  9) + (1 <<  8), 1},
	{ 1 << 10, 1 }, { (1 << 10) + (1 <<  9), 1},
	{ 1 << 11, 1 }, { (1 << 11) + (1 << 10), 1},
	{ 1 << 12, 0 }, { (1 << 12) + (1 << 11), 1},
	{ 1 << 13, 1 }, { (1 << 13) + (1 << 12), 1},
	{ 1 << 14, 1 }, { (1 << 14) + (1 << 13), 1},
	{ 1 << 15, 1 }, { (1 << 15) + (1 << 14), 1},
	{ 1 << 16, 0 }, { (1 << 16) + (1 << 15), 1},
	{ 1 << 17, 1 }, { (1 << 17) + (1 << 16), 1},
	{ 1 << 18, 1 }, { (1 << 18) + (1 << 17), 1},
	{ 1 << 19, 1 }, { (1 << 19) + (1 << 18), 1},
	{ 1 << 20, 0 }, { (1 << 20) + (1 << 19), 1},
	{ 1 << 21, 1 }, { (1 << 21) + (1 << 20), 1},
	{ 1 << 22, 1 }, { (1 << 22) + (1 << 21), 1},
};
#define TEST_CNT (sizeof test_size / sizeof test_size[0])

static int rs, lrs;
static int use_async;
static int use_rgai;
static int verify;
static int flags;
static int custom;
static int use_fork;
static pid_t fork_pid;
static int size_option;
static int done;
static int iterations = 1;
static int transfer_size = 1000;
static int transfer_count = 1000;
static long buffer_size = -1;
static char test_name[10] = "custom";
static char *port = "7471";
static int keepalive;
static char *dst_addr;
static char *src_addr;
static struct timeval start, end;
static struct timeval user_usage, kernel_usage;
static struct rusage start_usage;
static void **buf;
static exs_mhandle_t *buf_mhandle;
static exs_qhandle_t qhandle;
static struct rdma_addrinfo rai_hints;
static struct addrinfo ai_hints;

static void show_perf(void)
{
	char str[32];
	float usec;
	long long bytes;

	usec = (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec);
	bytes = (long long) iterations * transfer_count * transfer_size;
	if (!use_async) {
		bytes *= 2;
	}

	/* name size transfers iterations bytes seconds Gb/sec usec/xfer */
	printf("%-10s", test_name);
	size_str(str, sizeof str, transfer_size);
	printf("%-8s", str);
	cnt_str(str, sizeof str, transfer_count);
	printf("%-8s", str);
	cnt_str(str, sizeof str, iterations);
	printf("%-8s", str);
	size_str(str, sizeof str, bytes);
	printf("%-8s", str);
	printf("%8.2fs%10.2f%11.2f\n",
		usec / 1000000., (bytes * 8) / (1000. * usec),
		(usec / iterations) / (transfer_count * 2));

	fprintf(stderr, "%11ld.%06ld seconds user\n",
			user_usage.tv_sec, user_usage.tv_usec);
	fprintf(stderr, "%11ld.%06ld seconds kernel\n",
			kernel_usage.tv_sec, kernel_usage.tv_usec);
}	/* show_perf */

static void init_latency_test(int size)
{
	char sstr[5];

	size_str(sstr, sizeof sstr, size);
	snprintf(test_name, sizeof test_name, "%s_lat", sstr);
	transfer_count = 1;
	transfer_size = size;
	iterations = size_to_count(transfer_size);
}	/* init_latency_test */

static void init_bandwidth_test(int size)
{
	char sstr[5];

	size_str(sstr, sizeof sstr, size);
	snprintf(test_name, sizeof test_name, "%s_bw", sstr);
	iterations = 1;
	transfer_size = size;
	transfer_count = size_to_count(transfer_size);
}	/* init_bandwidth_test */

static int wait_for_event(struct exs_event *evt)
{
	int ret;

	while ((ret = exs_qdequeue(qhandle, evt, 1, NULL)) > 0) {
		if (evt->exs_evt_errno != 0) {
			ret = -1;
			errno = evt->exs_evt_errno;
			break;
		} else {
			ret = evt->exs_evt_union.exs_evt_xfer.exs_evt_length;
			break;
		}
	}

	return ret;
}	/* wait_for_event */


static int send_xfer(int bufno, int size)
{
	int offset, ret;

	assert(bufno < transfer_count);
	if (verify)
		format_buf(buf[bufno], size);

	for (offset = 0; offset < size; ) {
		assert(bufno < transfer_count);
		ret = exs_blocking_send(rs, buf[bufno] + offset, size - offset, flags,
				//qhandle, (void *)(uintptr_t)offset,
				buf_mhandle[bufno]);
		if (ret > 0) {
			offset += ret;
		} else if (errno != EWOULDBLOCK && errno != EAGAIN) {
			perror("exs_send");
			fprintf(stderr, "buf %p length %zu flags %d mhandle %p bufno %d\n",
					buf[bufno] + offset, (size_t)(size - offset), flags,
					buf_mhandle[bufno], bufno);
			return ret;
		}
	}

	return 0;
}	/* send_xfer */

static int recv_xfer(int bufno, int size)
{
	int offset, ret;

	for (offset = 0; offset < size; ) {
		ret = exs_blocking_recv(rs, buf[bufno] + offset, size - offset, flags,
				buf_mhandle[bufno]);
		if (ret > 0) {
			offset += ret;
		} else if (ret == 0) {
			fprintf(stderr, "Got unexpected EOF on exs_recv\n");
			return ret;
		} else if (errno != EWOULDBLOCK && errno != EAGAIN) {
			perror("exs_recv");
			return ret;
		}
	}

	if (verify) {
		ret = verify_buf(buf[bufno], size);
		if (ret)
			return ret;
	}

	return 0;
}	/* recv_xfer */

static int prime_async(void)
{
	int ret = 0;
	int t;

	for (t = 0; !ret && t < transfer_count; t++) {
		ret = exs_recv(rs, buf[t], transfer_size, flags | MSG_WAITALL,
			qhandle, (void *)(uintptr_t)t,
			buf_mhandle[t]);
		if (ret < 0) {
			perror("exs_recv start");
		}
	}

	return ret;
}	/* prime_async */

static int sync_test(void)
{
	int ret;

	ret = dst_addr ? send_xfer(0, 16) : recv_xfer(0, 16);
	if (ret)
		return ret;

	if (!done && !dst_addr && use_async) {
		prime_async();
	}

	return dst_addr ? recv_xfer(0, 16) : send_xfer(0, 16);
}	/* sync_test */

static int run_async_test_server(void)
{
	struct exs_event evt;
	int rem;
	int ret, fail, i;
	unsigned int t;

	fail = 0;
	i = 0;
	t = 0;
	rem = transfer_count;
	while (!fail && rem > 0) {
		ret = wait_for_event(&evt);
		switch (evt.exs_evt_type) {
		case EXS_EVT_SEND:
			assert(ret < 0);
			perror("exs event completion");
			fail = -1;
			rem--;
			goto out;
		case EXS_EVT_RECV:
			if (ret < 0) {
				perror("exs event completion");
				fail = -1;
				rem--;
				goto out;
			} else if (verify) {
				verify_buf(evt.exs_evt_union.exs_evt_xfer.exs_evt_buffer,
					evt.exs_evt_union.exs_evt_xfer.exs_evt_length);
			}
			assert((uintptr_t)evt.exs_evt_ahandle == t);
			assert(evt.exs_evt_union.exs_evt_xfer.exs_evt_length == (size_t)transfer_size);
			if (t == 0) {
				i++;
			}
			if (!fail && i < iterations) {
				ret = exs_recv(rs, buf[t],
					transfer_size, flags | MSG_WAITALL,
					qhandle, (void *)(uintptr_t)t, buf_mhandle[t]);
				if (ret < 0) {
					perror("exs_recv start");
					fail = -1;
				}
			} else {
				rem--;
			}
			ret = exs_send(rs, buf[t], transfer_size, flags | EXS_UNSIGNALED,
				qhandle, (void *)(uintptr_t)t,
				buf_mhandle[t]);
			if (ret < 0) {
				perror("exs_send start");
				fail = -1;
			}
			t = (t + 1) % transfer_count;
			break;
		default:
			fprintf(stderr, "Got unexpected event type %s\n",
				exs_evt_type_to_string(evt.exs_evt_type));
		}
	}
out:
	return fail;
}	/* run_async_test_server */

static int run_async_test_client(void)
{
	struct exs_event evt;
	int rem;
	int ret, fail, i;
	unsigned int t;

	fail = 0;
	rem = 0;
	i = 0;
	prime_async();
	for (t = 0; !fail && t < (unsigned int)transfer_count; t++) {
		if (verify) {
			format_buf(buf[t], transfer_size);
		}
#if 0
		ret = exs_recv(rs, buf[t], transfer_size, flags | MSG_WAITALL,
			qhandle, (void *)(uintptr_t)t,
			buf_mhandle[t]);
		if (ret < 0) {
			perror("exs_recv start");
			fail = -1;
			break;
		}
#endif
		ret = exs_send(rs, buf[t], transfer_size, flags | EXS_UNSIGNALED,
			qhandle, (void *)(uintptr_t)t,
			buf_mhandle[t]);
		if (ret < 0) {
			perror("exs_send start");
			fail = -1;
			break;
		}
		rem++;
	}
	t = 0;
	while (!fail && rem > 0) {
		ret = wait_for_event(&evt);
		switch (evt.exs_evt_type) {
		case EXS_EVT_RECV:
			if (ret < 0) {
				perror("exs_recv completion");
				fail = -1;
				goto out;
			} else if (verify) {
				verify_buf(evt.exs_evt_union.exs_evt_xfer.exs_evt_buffer,
					evt.exs_evt_union.exs_evt_xfer.exs_evt_length);
			}
			assert(t == (uintptr_t)evt.exs_evt_ahandle);
			if (t == 0) {
				i++;
			}
			if (!fail && i < iterations) {
				if (verify) {
					format_buf(buf[t], transfer_size);
				}
				ret = exs_recv(rs, buf[t], transfer_size, flags | MSG_WAITALL,
					qhandle, (void *)(uintptr_t)t,
					buf_mhandle[t]);
				if (ret < 0) {
					perror("exs_recv start");
					fail = -1;
				}
				ret = exs_send(rs, buf[t], transfer_size, flags | EXS_UNSIGNALED,
					qhandle, (void *)(uintptr_t)t,
					buf_mhandle[t]);
				if (ret < 0) {
					perror("exs_send start");
					fail = -1;
				}
			} else {
				rem--;
			}
			t = (t + 1) % transfer_count;
			break;
		case EXS_EVT_SEND:
			/* Can only get EXS_EVT_SEND in case of error */
			assert(ret < 0);
			perror("exs_send completion");
			fail = -1;
			goto out;
		default:
			fprintf(stderr, "Got unexpected event type %s\n",
				exs_evt_type_to_string(evt.exs_evt_type));
		}
	}
out:
	return fail;
}	/* run_async_test_client */

static int run_test(void)
{
	int ret, i, t;

	done = 0;
	ret = sync_test();
	if (ret)
		goto out;

	if (getrusage(RUSAGE_SELF, &start_usage) < 0) {
		perror("getrusage start");
		goto out;
	}
	gettimeofday(&start, NULL);
	if (!use_async) {
		for (i = 0; i < iterations; i++) {
			for (t = 0; t < transfer_count; t++) {
				ret = dst_addr ? send_xfer(t, transfer_size) :
						 recv_xfer(t, transfer_size);
				if (ret)
					goto out;
			}

			for (t = 0; t < transfer_count; t++) {
				ret = dst_addr ? recv_xfer(t, transfer_size) :
						 send_xfer(t, transfer_size);
				if (ret)
					goto out;
			}
		}
	} else {
		ret = dst_addr ? run_async_test_client() : run_async_test_server();
		if (ret)
			goto out;
	}
	gettimeofday(&end, NULL);
	{
		int us, uu, ks, ku;
		get_relative_usage(&start_usage, &us, &uu, &ks, &ku, __func__);
		user_usage.tv_sec = us;
		user_usage.tv_usec = uu;
		kernel_usage.tv_sec = ks;
		kernel_usage.tv_usec = ku;
	}

	/* Do one final exchange to ensure all data is transmitted.
	 * FIXME: This should be fixed in UNH EXS so that this step is
	 * unnecessary. */
	done = 1;
	ret = sync_test();
	if (ret)
		goto out;

	show_perf();
	ret = 0;

out:
	return ret;
}	/* run_test */

static void set_options(int rs)
{
	long val;

	if (buffer_size >= 0) {
		if (exs_fcntl(rs, EXS_F_SETSTREAMBUFSIZE, buffer_size) < 0) {
			perror("exs_fcntl EXS_F_SETSTREAMBUFSIZE");
		}
	} else {
		/* Leave at largest possible size for best performance */
		/* exs_fcntl(rs, EXS_F_SETSTREAMBUFSIZE, 1 << 19); */
	}

#if 0
	val = exs_fcntl(rs, EXS_F_GETFD, 0);
	val |= EXS_FD_NO_STREAM_ADVERT;
	exs_fctnl(rs, EXS_F_SETFD, val);
#endif

	if (transfer_count > 0) {
		if (exs_fcntl(rs, EXS_F_SETFLOWCONTROLCREDITS, transfer_count) < 0) {
			perror("exs_fcntl EXS_F_SETFLOWCONTROLCREDITS");
		}
	}

	/* Inline size based on experimental data */
	val = 384;
	if (exs_fcntl(rs, EXS_F_SETINLINEMAXSIZE, val) < 0) {
		perror("exs_fcntl EXS_F_SETINLINEMAXSIZE");
	}
}	/* set_options */

static int server_listen(void)
{
	struct addrinfo *ai;
	int ret;

	ai_hints.ai_flags |= AI_PASSIVE;
	ret = getaddrinfo(src_addr, port, &ai_hints, &ai);
	if (ret) {
		perror("getaddrinfo");
		return ret;
	}

	lrs = exs_socket(ai->ai_family, SOCK_STREAM, 0);
	if (lrs < 0) {
		perror("exs_socket");
		ret = lrs;
		goto free;
	}

	ret = exs_bind(lrs, ai->ai_addr, ai->ai_addrlen);
	if (ret) {
		perror("rbind");
		goto close;
	}

	ret = exs_listen(lrs, 1);
	if (ret)
		perror("rlisten");

close:
	if (ret)
		exs_blocking_close(lrs);
free:
	freeaddrinfo(ai);
	return ret;
}	/* server_listen */

static int server_connect(void)
{
	int ret = 0;

	set_options(lrs);
	do {
		rs = exs_blocking_accept(lrs, NULL, 0);
	} while (rs < 0 && (errno == EAGAIN || errno == EWOULDBLOCK));
	if (rs < 0) {
		perror("raccept");
		return rs;
	}

	return ret;
}	/* server_connect */

static int client_connect(void)
{
	struct rdma_addrinfo *rai = NULL;
	struct addrinfo *ai;
	int ret;

	ret = getaddrinfo(dst_addr, port, &ai_hints, &ai);

	if (ret) {
		perror("getaddrinfo");
		return ret;
	}

	rs = exs_socket(ai->ai_family, SOCK_STREAM, 0);
	if (rs < 0) {
		perror("exs_socket");
		ret = rs;
		goto free;
	}

	set_options(rs);
	/* TODO: bind client to src_addr */

	ret = exs_blocking_connect(rs, ai->ai_addr, ai->ai_addrlen);
	if (ret && (errno != EINPROGRESS)) {
		perror("rconnect");
		goto close;
	}

close:
	if (ret)
		exs_blocking_close(rs);
free:
	if (rai)
		rdma_freeaddrinfo(rai);
	else
		freeaddrinfo(ai);
	return ret;
}	/* client_connect */

static int alloc_bufs(void)
{
	size_t buf_size;
	int j;

	buf_size = (transfer_size < 16) ? 16 : transfer_size;
	buf = calloc(transfer_count, sizeof(*buf));
	if (!buf) {
		perror("calloc");
		goto fail0;
	}
	buf_mhandle = malloc(transfer_count * sizeof(*buf_mhandle));
	if (!buf_mhandle) {
		perror("malloc");
		goto fail1;
	}
	memset(buf_mhandle, (uintptr_t)EXS_MHANDLE_INVALID, transfer_count);
	for (j = 0; j < transfer_count; j++) {
		posix_memalign(&buf[j], 64, buf_size);
		if (!buf) {
			perror("posix_memalign");
			goto fail;
		}

		buf_mhandle[j] = exs_mregister(buf[j], buf_size, EXS_ACCESS_ALL);
		if (buf_mhandle[j] == EXS_MHANDLE_INVALID) {
			perror("exs_mregister");
			free(buf[j]);
			goto fail;
		}
	}

	return 0;

fail:
	/* free anything we just set up */
	/* we assume that j is the buffer that failed */
	for (j--; j >= 0; j--) {
		exs_mderegister(buf_mhandle[j], 0);
		free(buf[j]);
	}
	free(buf_mhandle);
	buf_mhandle = NULL;
fail1:
	free(buf);
	buf = NULL;
fail0:
	return -1;
}	/* alloc_bufs */

static void dealloc_bufs(void)
{
	int j;

	if (buf) {
		for (j = 0; j < transfer_count; j++) {
			exs_mderegister(buf_mhandle[j], 0);
			free(buf[j]);
		}
	}
	free(buf_mhandle);
	buf_mhandle = NULL;
	free(buf);
	buf = NULL;
}	/* dealloc_bufs */

static int run(void)
{
	unsigned int i;
	int ret = 0;

	if (use_async) {
		qhandle = exs_qcreate(2 * transfer_count);
		if (qhandle == EXS_QHANDLE_INVALID) {
			goto free;
		}
	}

	if (!dst_addr) {
		ret = server_listen();
		if (ret)
			goto free;
	}

	printf("%-10s%-8s%-8s%-8s%-8s%8s %10s%13s\n",
	       "name", "bytes", "xfers", "iters", "total", "time", "Gb/sec", "usec/xfer");
	if (!custom) {
		ret = dst_addr ? client_connect() : server_connect();
		if (ret)
			goto free;

		for (i = 0; i < TEST_CNT && !fork_pid; i++) {
			if (test_size[i].option > size_option)
				continue;
			init_latency_test(test_size[i].size);
			if (alloc_bufs()) {
				ret = -1;
				goto free;
			}
			run_test();
			dealloc_bufs();
		}
		/*exs_blocking_shutdown(rs, SHUT_RDWR);*/
		exs_blocking_close(rs);

		if (!dst_addr && use_fork && !fork_pid)
			goto free;

		ret = dst_addr ? client_connect() : server_connect();
		if (ret)
			goto free;
		for (i = 0; i < TEST_CNT && !fork_pid; i++) {
			if (test_size[i].option > size_option)
				continue;
			init_bandwidth_test(test_size[i].size);
			if (alloc_bufs()) {
				goto free;
			}
			run_test();
			dealloc_bufs();
		}
	} else {
		if (alloc_bufs()) {
			goto free;
		}
		ret = dst_addr ? client_connect() : server_connect();
		if (ret)
			goto free;

		if (!fork_pid)
			ret = run_test();
		dealloc_bufs();
	}

	/*exs_blocking_shutdown(rs, SHUT_RDWR);*/
	exs_blocking_close(rs);
free:
	if (use_async) {
		if (qhandle != EXS_QHANDLE_INVALID && exs_qdelete(qhandle)) {
			perror("exs_qdelete");
		}
	}
	dealloc_bufs();
	return ret;
}	/* run */

static int set_test_opt(char *optarg)
{
	if (strlen(optarg) == 1) {
		switch (optarg[0]) {
		case 'a':
			use_async = 1;
			break;
		case 'b':
			flags = (flags & ~MSG_DONTWAIT);
			break;
		case 'v':
			verify = 1;
			break;
		default:
			return -1;
		}
	} else {
		if (!strncasecmp("async", optarg, 5)) {
			use_async = 1;
		} else if (!strncasecmp("block", optarg, 5)) {
			flags = (flags & ~MSG_DONTWAIT) | MSG_WAITALL;
		} else if (!strncasecmp("verify", optarg, 6)) {
			verify = 1;
		} else {
			return -1;
		}
	}
	return 0;
}	/* set_test_opt */

int main(int argc, char **argv)
{
	char *endptr;
	int op, ret;

	exs_init(1);
	ai_hints.ai_socktype = SOCK_STREAM;
	rai_hints.ai_port_space = RDMA_PS_TCP;
	while ((op = getopt(argc, argv, "s:b:f:B:I:C:S:p:k:T:")) != -1) {
		switch (op) {
		case 's':
			dst_addr = optarg;
			break;
		case 'b':
			src_addr = optarg;
			break;
		case 'f':
			if (!strncasecmp("ip", optarg, 2)) {
				ai_hints.ai_flags = AI_NUMERICHOST;
			} else if (!strncasecmp("gid", optarg, 3)) {
				rai_hints.ai_flags = RAI_NUMERICHOST | RAI_FAMILY;
				rai_hints.ai_family = AF_IB;
				use_rgai = 1;
			} else {
				fprintf(stderr, "Warning: unknown address format\n");
			}
			break;
		case 'B':
			errno = 0;
			buffer_size = strtol(optarg, &endptr, 0);
			if (errno != 0) {
				perror("strtol");
				exit(EXIT_FAILURE);
			} else if (endptr == optarg || *endptr != '\0') {
				fprintf(stderr, "invalid buffer size \"%s\"\n",
						optarg);
				exit(EXIT_FAILURE);
			}
			break;
		case 'I':
			custom = 1;
			iterations = atoi(optarg);
			break;
		case 'C':
			custom = 1;
			transfer_count = atoi(optarg);
			break;
		case 'S':
			if (!strncasecmp("all", optarg, 3)) {
				size_option = 1;
			} else {
				custom = 1;
				transfer_size = atoi(optarg);
			}
			break;
		case 'p':
			port = optarg;
			break;
		case 'k':
			keepalive = atoi(optarg);
			break;
		case 'T':
			if (!set_test_opt(optarg))
				break;
			/* invalid option - fall through */
		default:
			printf("usage: %s\n", argv[0]);
			printf("\t[-s server_address]\n");
			printf("\t[-b bind_address]\n");
			printf("\t[-f address_format]\n");
			printf("\t    name, ip, ipv6, or gid\n");
			printf("\t[-B buffer_size]\n");
			printf("\t[-I iterations]\n");
			printf("\t[-C transfer_count]\n");
			printf("\t[-S transfer_size or all]\n");
			printf("\t[-p port_number]\n");
			printf("\t[-k keepalive_time]\n");
			printf("\t[-T test_option]\n");
			printf("\t    a|async - asynchronous operation\n");
			printf("\t    b|blocking - use exs_blocking calls\n");
			printf("\t    v|verify - verify data\n");
			exit(1);
		}
	}

	ret = run();
	exs_dump_counters();
	return ret;
}	/* main */
